unit hiScrollBarEx;

interface

uses
  Kol,Windows,Share,Debug,Win;

type
  TATScrollElemType = (
    aseArrowUp,
    aseArrowDown,
    aseArrowLeft,
    aseArrowRight,
    aseScrollThumbV,
    aseScrollThumbH,
    aseScrollAreaH,
    aseScrollAreaV,
    aseScrolledAreaH,
    aseScrolledAreaV,
    aseCorner
    );

  TATScrollDrawEvent = procedure (Sender: TObject; AType: TATScrollElemType;
    ACanvas: PCanvas; const ARect: TRect; var ACanDraw: boolean) of object;

  TScrollBarKind = (Horizontal, Vertical);

  THIScrollBarEx = class(THIWin)
  private
    FKind: TScrollBarKind;
    FTimerDelay: Integer;
    FPos,
    FMin,
    FMax,
    FPage: Integer;
    FAutoSize: Boolean;

    //internal
    FIn: TRect; //area for scrolling
    FInUp: TRect; //area for up or left arrow
    FInDown: TRect; //area for down or right arrow
    FInThumb: TRect; //area for scroll-thumb
    FInCorner: TRect;
    FInPageUp: TRect;
    FInPageDown: TRect;
    FInit: boolean;
    FBitmap: Kol.PBitmap;
    FTimer: PTimer;

    FOnOwnerDraw: TATScrollDrawEvent;

    //drag-drop
    FMouseDown: boolean;
    FMouseDragOffset: Integer;
    FMouseDownOnUp,
    FMouseDownOnDown,
    FMouseDownOnThumb,
    FMouseDownOnPageUp,
    FMouseDownOnPageDown: boolean;

    procedure DoPaintArrow(C: PCanvas; const R: TRect; Typ: TATScrollElemType);
    procedure DoPaintThumb(C: PCanvas);
    procedure DoPaintBack(C: PCanvas);
    procedure DoPaintBackScrolled(C: PCanvas);
    procedure DoPaintTo(C: PCanvas);

    procedure DoPaintStd_Corner(C: PCanvas; const R: TRect);
    procedure DoPaintStd_Back(C: PCanvas; const R: TRect);
    procedure DoPaintStd_BackScrolled(C: PCanvas; const R: TRect);
    procedure DoPaintStd_Arrow(C: PCanvas; R: TRect; Typ: TATScrollElemType);
    procedure DoPaintStd_Thumb(C: PCanvas; const R: TRect);

    function IsHorz: boolean;
    function MouseToPos(X, Y: Integer): Integer;
    procedure DoUpdateThumbRect;
    procedure DoUpdateCornerRect;
    procedure DoUpdatePosOnDrag(X, Y: Integer);
    procedure DoScrollBy(NDelta: Integer);
    function GetPxAtScroll(APos: Integer): Integer;

    procedure onTimer(Sender: PObj);
    procedure SetAutoSize(Value: boolean);
    procedure SetKind(Value: TScrollBarKind);
    procedure SetPage(Value: Integer);
    procedure SetPos(Value: Integer);
    procedure SetMin(Value: Integer);
    procedure SetMax(Value: Integer);
    function DoDrawEvent(AType: TATScrollElemType;
      ACanvas: PCanvas; const ARect: TRect): boolean;
    procedure Paint(Sender: PControl; DC: HDC);
    procedure Resize(Obj:PObj);
    procedure onMouseDown(Sender: PControl; var Mouse: TMouseEventData);
    procedure onMouseUp(Sender: PControl; var Mouse: TMouseEventData);
    procedure onMouseMove(Sender: PControl; var Mouse: TMouseEventData);
    procedure onMouseLeave( Sender: PObj );
  public
    _prop_IndentBorder: Integer;
    _prop_IndentCorner: Integer;
    _prop_IndentArrow: Integer;
    _prop_IndentArrLonger: Integer;
    _prop_ColorBorder: TColor;
    _prop_ColorFill: TColor;
    _prop_ColorScrolled: TColor;
    _prop_ColorRect: TColor;
    _prop_ColorArrow: TColor;
    _prop_ScrollMode: byte;
    _event_onPosition:THI_Event;

    procedure Init;
    destructor Destroy; override;
    procedure _work_doPageSize(var _Data: TData; Index: word);
    procedure _work_doPosition(var _Data: TData; Index: word);
    procedure _work_doMin(var _Data: TData; Index: word);
    procedure _work_doMax(var _Data: TData; Index: word);
    procedure _var_Position(var _Data: TData; Index: word);

    property TimerDelay: Integer read FTimerDelay write FTimerDelay;
    property _prop_AutoSize: boolean read FAutoSize write SetAutoSize;
    property _prop_PageSize: Integer read FPage write SetPage;
    property _prop_Position: Integer read FPos write SetPos;
    property _prop_Max: Integer read FMax write SetMax;
    property _prop_Min: Integer read FMin write SetMin;
    property _prop_Kind: TScrollBarKind read FKind write SetKind;

    property OnOwnerDraw: TATScrollDrawEvent read FOnOwnerDraw write FOnOwnerDraw;
  end;

implementation

function CenterPoint(const Rect: TRect): TPoint;
begin
  with Rect do
  begin
    Result.X := (Right - Left) div 2 + Left;
    Result.Y := (Bottom - Top) div 2 + Top;
  end;
end;

{ THIScrollBarEx }
procedure THIScrollBarEx._work_doPosition;
begin
  _prop_Position := ToInteger(_Data);
  Control.Invalidate;
end;

procedure THIScrollBarEx._work_doMin;
begin
  _prop_Min := ToInteger(_Data);
  Control.Invalidate;
end;

procedure THIScrollBarEx._work_doMax;
begin
  _prop_Max := ToInteger(_Data);
  Control.Invalidate;
end;

procedure THIScrollBarEx._work_doPageSize;
begin
  _prop_PageSize := ToInteger(_Data);
  Control.Invalidate;
end;

procedure THIScrollBarEx._var_Position;
begin
  dtInteger(_Data, FPos);
end;

procedure THIScrollBarEx.Init;
begin
  Control := NewPanel(FParent,esNone);
  Control.Caption:= '';
  inherited;
  FKind:= TScrollBarKind(_prop_Kind);

  FBitmap:= NewBitmap(1, 1);

  FTimerDelay:= 80;
  FTimer:= NewTimer(FTimerDelay);
  FTimer.Enabled:= False;
  FTimer.OnTimer:= onTimer;
  FInit := True;

  FMouseDown:= False;
  FMouseDragOffset:= 0;
  Control.OnMouseDown := onMouseDown;
  Control.OnMouseUp := onMouseUp;
  Control.OnMouseMove := onMouseMove;
  Control.onMouseLeave := onMouseLeave;
  Control.OnResize := Resize;
  Control.OnPaint := Paint;
  Resize(Control);
end;

destructor THIScrollBarEx.Destroy;
begin
  FTimer.Enabled:= False;
  FBitmap.Free;
  inherited;
end;

procedure THIScrollBarEx.Paint;
begin
  if Assigned(FBitmap) then
  begin
    DoPaintTo(FBitmap.Canvas);
    Control.Canvas.CopyRect(Control.ClientRect, FBitmap.Canvas, Control.ClientRect);
  end;
end;

procedure THIScrollBarEx.DoPaintTo(C: PCanvas);
var
  fSize: Integer;
begin
  FIn:= Control.ClientRect;

  DoUpdateCornerRect;
  if not IsRectEmpty(FInCorner) then
    if DoDrawEvent(aseCorner, C, FInCorner) then
      DoPaintStd_Corner(C, FInCorner);

  C.Brush.Color:= _prop_ColorBorder;
  C.FillRect(FIn);

  FIn:= MakeRect(
    FIn.Left+_prop_IndentBorder,
    FIn.Top+_prop_IndentBorder,
    FIn.Right-_prop_IndentBorder,
    FIn.Bottom-_prop_IndentBorder);

  if IsHorz then
  begin
    FSize:= Kol.Min(FIn.Bottom-FIn.Top, Round((FIn.Right-FIn.Left) div 2));
    Inc(FSize, _prop_IndentArrLonger);
    FInUp:= MakeRect(FIn.Left, FIn.Top, FIn.Left+FSize, FIn.Bottom);
    FInDown:= MakeRect(FIn.Right-FSize, FIn.Top, FIn.Right, FIn.Bottom);
    DoPaintArrow(C, FInUp, aseArrowLeft);
    DoPaintArrow(C, FInDown, aseArrowRight);
    Inc(FIn.Left, FSize);
    Dec(FIn.Right, FSize);
  end
  else
  begin
    FSize:= Kol.Min(FIn.Right-FIn.Left, Round((FIn.Bottom-FIn.Top) div 2));
    Inc(FSize, _prop_IndentArrLonger);
    FInUp:= MakeRect(FIn.Left, FIn.Top, FIn.Right, FIn.Top+FSize);
    FInDown:= MakeRect(FIn.Left, FIn.Bottom-FSize, FIn.Right, FIn.Bottom);
    DoPaintArrow(C, FInUp, aseArrowUp);
    DoPaintArrow(C, FInDown, aseArrowDown);
    Inc(FIn.Top, FSize);
    Dec(FIn.Bottom, FSize);
  end;

  DoPaintBack(C);
  DoUpdateThumbRect;
  DoPaintBackScrolled(C);
  DoPaintThumb(C);
end;

procedure THIScrollBarEx.DoPaintBack(C: PCanvas);
var
  Typ: TATScrollElemType;
begin
  if IsHorz then Typ:= aseScrollAreaH else Typ:= aseScrollAreaV;
  if DoDrawEvent(Typ, C, FIn) then
    DoPaintStd_Back(C, FIn);
end;

procedure THIScrollBarEx.DoPaintBackScrolled(C: PCanvas);
var
  Typ: TATScrollElemType;
begin
  if IsHorz then Typ:= aseScrolledAreaH else Typ:= aseScrolledAreaV;

  if FMouseDown and FMouseDownOnPageUp then
    if DoDrawEvent(Typ, C, FInPageUp) then
      DoPaintStd_BackScrolled(C, FInPageUp);

  if FMouseDown and FMouseDownOnPageDown then
    if DoDrawEvent(Typ, C, FInPageDown) then
      DoPaintStd_BackScrolled(C, FInPageDown);
end;

procedure THIScrollBarEx.onMouseDown(Sender: PControl; var Mouse: TMouseEventData);
begin
  with Mouse do
  begin
    FMouseDown:= Button=mbLeft;
    FMouseDownOnThumb:= PtInRect(FInThumb, MakePoint(X, Y));
    FMouseDownOnUp:= PtInRect(FInUp, MakePoint(X, Y));
    FMouseDownOnDown:= PtInRect(FInDown, MakePoint(X, Y));
    FMouseDownOnPageUp:= PtInRect(FInPageUp, MakePoint(X, Y));
    FMouseDownOnPageDown:= PtInRect(FInPageDown, MakePoint(X, Y));

    if IsHorz then
      FMouseDragOffset:= X-FInThumb.Left
    else
      FMouseDragOffset:= Y-FInThumb.Top;

    FTimer.Enabled:= FMouseDown and
      (FMouseDownOnUp or
       FMouseDownOnDown or
       FMouseDownOnPageUp or
       FMouseDownOnPageDown);
  end;
end;

procedure THIScrollBarEx.onMouseUp(Sender: PControl; var Mouse: TMouseEventData);
begin
  FMouseDown:= False;
  FMouseDownOnThumb:= False;
  FTimer.Enabled:= False;
  Control.Invalidate;
  if _prop_ScrollMode = 1 then
    _hi_OnEvent(_event_onPosition, FPos);
end;

procedure THIScrollBarEx.onMouseLeave;
const Pos: Integer = 0;
begin
  FMouseDown:= False;
  FMouseDownOnThumb:= False;
  FTimer.Enabled:= False;
  if (_prop_ScrollMode = 1) and (Pos <> FPos) then
    _hi_OnEvent(_event_onPosition, FPos);
  Control.Invalidate;
  Pos := FPos;
end;

procedure THIScrollBarEx.Resize;
begin
  if Assigned(FBitmap) then
  begin
    //little complicated to speed up
    if IsHorz then
    begin
      FBitmap.Width:= Kol.Max(FBitmap.Width, Control.Width);
      FBitmap.Height:= Control.Height;
    end
    else
    begin
      FBitmap.Width:= Control.Width;
      FBitmap.Height:= Kol.Max(FBitmap.Height, Control.Height);
    end;
  end;
  Control.Invalidate;
end;

function THIScrollBarEx.DoDrawEvent(AType: TATScrollElemType;
  ACanvas: PCanvas; const ARect: TRect): boolean;
begin
  Result:= True;
  if Assigned(FOnOwnerDraw) then
    FOnOwnerDraw(Self, AType, ACanvas, ARect, Result);
end;

procedure THIScrollBarEx.SetKind(Value: TScrollBarKind);
begin
  if Value<>FKind then
  begin
    FKind:= Value;
    //Control.Invalidate;
  end;
end;

procedure THIScrollBarEx.DoPaintArrow(C: PCanvas; const R: TRect;
  Typ: TATScrollElemType);
begin
  if DoDrawEvent(Typ, C, R) then
    DoPaintStd_Arrow(C, R, Typ);
end;

procedure THIScrollBarEx.DoPaintStd_Arrow(C: PCanvas; R: TRect;
  Typ: TATScrollElemType);
var
  P, P1, P2, P3: TPoint;
  cc: Integer;
begin
  C.Brush.Color:= _prop_ColorRect;
  C.FillRect(R);

  InflateRect(R, -1, -1);
  C.Brush.Color:= _prop_ColorFill;
  C.FillRect(R);

  P:= CenterPoint(R);
  cc:= _prop_IndentArrow;

  case Typ of
    aseArrowUp:
      begin
        P1:= MakePoint(P.X-cc, P.Y+cc div 2);
        P2:= MakePoint(P.X+cc, P.Y+cc div 2);
        P3:= MakePoint(P.X, P.Y-cc+cc div 2);
      end;
    aseArrowDown:
      begin
        P1:= MakePoint(P.X-cc, P.Y-cc div 2);
        P2:= MakePoint(P.X+cc, P.Y-cc div 2);
        P3:= MakePoint(P.X, P.Y+cc-cc div 2);
      end;
    aseArrowLeft:
      begin
        P1:= MakePoint(P.X+cc div 2, P.Y-cc);
        P2:= MakePoint(P.X+cc div 2, P.Y+cc);
        P3:= MakePoint(P.X-cc+cc div 2, P.Y);
      end;
    aseArrowRight:
      begin
        P1:= MakePoint(P.X-cc div 2    -1, P.Y-cc);
        P2:= MakePoint(P.X-cc div 2    -1, P.Y+cc);
        P3:= MakePoint(P.X+cc-cc div 2 -1, P.Y);
      end;
    else
      Exit;
 end;

  C.Brush.Color:= _prop_ColorArrow;
  C.Pen.Color:= _prop_ColorArrow;
  C.Polygon([P1, P2, P3]);
end;

function THIScrollBarEx.IsHorz: boolean;
begin
  Result:= FKind=TScrollBarKind(Horizontal);
end;

function THIScrollBarEx.GetPxAtScroll(APos: Integer): Integer;
var
  N0, NLen: Integer;
begin
  if IsHorz then
  begin
    N0:= FIn.Left;
    NLen:= FIn.Right-FIn.Left
  end
  else
  begin
    N0:= FIn.Top;
    NLen:= FIn.Bottom-FIn.Top;
  end;
  Result:= N0 + (APos-FMin) * NLen div (FMax-FMin);
end;

procedure THIScrollBarEx.DoUpdateThumbRect;
const
  cMinView = 10;
var
  R: TRect;
begin
  FInThumb:= MakeRect(0, 0, 0, 0);
  FInPageUp:= MakeRect(0, 0, 0, 0);
  FInPageDown:= MakeRect(0, 0, 0, 0);

  if IsHorz then
  begin
    if FIn.Right-FIn.Left<cMinView then Exit;
    R.Top:= FIn.Top;
    R.Bottom:= FIn.Bottom;
    R.Left:= GetPxAtScroll(FPos);
    R.Left:= Kol.Min(R.Left, FIn.Right-cMinView);
    R.Right:= GetPxAtScroll(FPos+FPage);
    R.Right:= Kol.Max(R.Right, R.Left+cMinView);
    R.Right:= Kol.Min(R.Right, FIn.Right);
  end
  else
  begin
    if FIn.Bottom-FIn.Top<cMinView then Exit;
    R.Left:= FIn.Left;
    R.Right:= FIn.Right;
    R.Top:= GetPxAtScroll(FPos);
    R.Top:= Kol.Min(R.Top, FIn.Bottom-cMinView);
    R.Bottom:= GetPxAtScroll(FPos+FPage);
    R.Bottom:= Kol.Max(R.Bottom, R.Top+cMinView);
    R.Bottom:=Kol.Min(R.Bottom, FIn.Bottom);
  end;
  FInThumb:= R;

  if IsHorz then
  begin
    FInPageUp:= MakeRect(FIn.Left, FIn.Top, FInThumb.Left, FIn.Bottom);
    FInPageDown:= MakeRect(FInThumb.Right, FIn.Top, FIn.Right, FIn.Bottom);
  end
  else
  begin
    FInPageUp:= MakeRect(FIn.Left, FIn.Top, FIn.Right, FInThumb.Top);
    FInPageDown:= MakeRect(FIn.Left, FInThumb.Bottom, FIn.Right, FIn.Bottom);
  end;
end;

procedure THIScrollBarEx.DoPaintThumb(C: PCanvas);
var
  Typ: TATScrollElemType;
begin
  if IsRectEmpty(FInThumb) then Exit;
  if IsHorz then
    Typ:= aseScrollThumbH
  else
    Typ:= aseScrollThumbV;

  if DoDrawEvent(Typ, C, FInThumb) then
    DoPaintStd_Thumb(C, FInThumb);
end;

procedure THIScrollBarEx.DoPaintStd_Thumb(C: PCanvas; const R: TRect);
const
  cMinMark = 20; //minimial size of thumb, after which thumb disappears
  cMarkOf = 4; //offset from thumb edge to "|||" lines
var
  P: TPoint;
begin
  C.Brush.Color:= _prop_ColorFill;
  C.Pen.Color:= _prop_ColorRect;
  C.Rectangle(R.Left,R.Top,R.Right,R.Bottom);

  P:= CenterPoint(R);
  if IsHorz then
  begin
    if (R.Right-R.Left)>cMinMark then
    begin
      C.MoveTo(P.X  , R.Top+cMarkOf);
      C.LineTo(P.X  , R.Bottom-cMarkOf);
      C.MoveTo(P.X-2, R.Top+cMarkOf);
      C.LineTo(P.X-2, R.Bottom-cMarkOf);
      C.MoveTo(P.X+2, R.Top+cMarkOf);
      C.LineTo(P.X+2, R.Bottom-cMarkOf);
    end;
  end
  else
  begin
    if (R.Bottom-R.Top)>cMinMark then
    begin
      C.MoveTo(R.Left+cMarkOf, P.Y);
      C.LineTo(R.Right-cMarkOf, P.Y);
      C.MoveTo(R.Left+cMarkOf, P.Y-2);
      C.LineTo(R.Right-cMarkOf, P.Y-2);
      C.MoveTo(R.Left+cMarkOf, P.Y+2);
      C.LineTo(R.Right-cMarkOf, P.Y+2);
    end;
  end;
end;

procedure THIScrollBarEx.SetAutoSize(Value: boolean);
begin
  FAutoSize := Value;
end;

procedure THIScrollBarEx.SetMax(Value: Integer);
begin
  if not _prop_AutoSize then
    begin
      if FMax<>(Value+FPage) then
      begin
        FMax:= Value+FPage;
        FPos:= Kol.Min(FPos, FMax);
      end;
    end else begin
      _prop_PageSize := Round(Value/15);
      FMax:= Value+_prop_PageSize;
      FPos:= Kol.Min(FPos, FMax);
    end;
  if FInit then
    Control.Invalidate;
end;

procedure THIScrollBarEx.SetMin(Value: Integer);
begin
  if FMin<>Value then
  begin
    FMin:= Value;
    FPos:= Kol.Max(FPos, FMin);
    if FInit then
      Control.Invalidate;
  end;
end;

procedure THIScrollBarEx.SetPage(Value: Integer);
begin
  if FPage<>Value then
  begin
    FPage:= Value;
    FPos:= Kol.Min(FPos, FMax);
    if FInit then
      Control.Invalidate;
  end;
end;

procedure THIScrollBarEx.SetPos(Value: Integer);
begin
  Value:= Kol.Min(Value, FMax);
  Value:= Kol.Max(Value, FMin);
  if FPos<>Value then
  begin
    FPos:= Value;
    Control.Invalidate;
    if _prop_ScrollMode = 0 then
      _hi_OnEvent(_event_onPosition, FPos);
  end;
end;

procedure THIScrollBarEx.onMouseMove(Sender: PControl; var Mouse: TMouseEventData);
begin
  inherited;

  if FMouseDownOnThumb then
  begin
    DoUpdatePosOnDrag(Mouse.X, Mouse.Y);
    Exit
  end;
end;

function THIScrollBarEx.MouseToPos(X, Y: Integer): Integer;
begin
  if IsHorz then
    Result:= FMin + (X-FIn.Left) * (FMax-FMin) div Kol.Max(FIn.Right-FIn.Left, 1)
  else
    Result:= FMin + (Y-FIn.Top) * (FMax-FMin) div Kol.Max(FIn.Bottom-FIn.Top, 1);
end;

procedure THIScrollBarEx.DoUpdatePosOnDrag(X, Y: Integer);
var
  N: Integer;
begin
  N:= MouseToPos(
    X-FMouseDragOffset,
    Y-FMouseDragOffset);
  N:= Kol.Max(N, FMin);
  N:= Kol.Min(N, FMax-FPage);
  SetPos(N);
end;

procedure THIScrollBarEx.DoScrollBy(NDelta: Integer);
var
  N: Integer;
begin
  N:= FPos;
  Inc(N, NDelta);
  if (NDelta>0) then
    N:= Kol.Min(N, FMax-FPage);
  SetPos(N);
end;

procedure THIScrollBarEx.onTimer(Sender: PObj);
var
  P: TPoint;
begin
  GetCursorPos(P);
  ScreenToClient(Control.Handle,P);

  if FMouseDownOnDown and PtInRect(FInDown, P) then
    DoScrollBy(1)
  else
  if FMouseDownOnUp and PtInRect(FInUp, P) then
    DoScrollBy(-1)
  else
  if FMouseDownOnPageDown and PtInRect(FInPageDown, P) then
    DoScrollBy(FPage)
  else
  if FMouseDownOnPageUp and PtInRect(FInPageUp, P) then
    DoScrollBy(-FPage);
end;

procedure THIScrollBarEx.DoPaintStd_Corner(C: PCanvas; const R: TRect);
begin
  C.Brush.Color:= Control.Color;
  C.FillRect(R);
end;

procedure THIScrollBarEx.DoPaintStd_Back(C: PCanvas; const R: TRect);
begin
  C.Brush.Color:= Control.Color;
  C.FillRect(R);
end;

procedure THIScrollBarEx.DoPaintStd_BackScrolled(C: PCanvas; const R: TRect);
begin
  C.Brush.Color:= _prop_ColorScrolled;
  C.FillRect(R);
end;

procedure THIScrollBarEx.DoUpdateCornerRect;
begin
  FInCorner:= MakeRect(0, 0, 0, 0);
  if IsHorz then
  begin
    if _prop_IndentCorner>0 then
    begin
      FInCorner:= MakeRect(Control.ClientWidth-_prop_IndentCorner, 0, Control.ClientWidth, Control.ClientHeight);
      Dec(FIn.Right, _prop_IndentCorner);
    end
    else
    if _prop_IndentCorner<0 then
    begin
      FInCorner:= MakeRect(0, 0, Abs(_prop_IndentCorner), Control.ClientHeight);
      Inc(FIn.Left, Abs(_prop_IndentCorner));
    end;
  end
  else
  begin
    if _prop_IndentCorner>0 then
    begin
      FInCorner:= MakeRect(0, Control.ClientHeight-_prop_IndentCorner, Control.ClientWidth, Control.ClientHeight);
      Dec(FIn.Bottom, _prop_IndentCorner);
    end
    else
    if _prop_IndentCorner<0 then
    begin
      FInCorner:= MakeRect(0, 0, Control.ClientWidth, Abs(_prop_IndentCorner));
      Inc(FIn.Top, Abs(_prop_IndentCorner));
    end;
  end;
end;

end.
